// Set up triangular from output from Triangle

#include<string>
#include<sstream>
#include<fstream>
#include<vector>
#include<set>
#include"intArray.h"

#include"../LinearAlgebra/svector.C"
#include"gelement.C"

typedef GeometricElement<2, 3> gelement;
typedef BoundaryGelement<2, 2> bgelement;

class triangleMesh
{
public:
  triangleMesh(std::string);                        // read mesh.poly file
  triangleMesh(const triangleMesh &mm) 
  {
    std::cerr << "triangleMesh(): don't copy a triangleMesh\n";
    throw("copying a triangleMesh");
  }
    
  triangleMesh() 
  {
    std::cerr << "triangleMesh(): Default constructor called, nyi\n";
    throw("bad triangleMesh constructor");
  }

  ~triangleMesh() {return;}
  
  int nNode()  const {return nodes;}
  int nElem()  const {return nelem;}
  int nBelem() const {return nbelem;}

  void info();
  void print();

  bool checkBdyOrientation();
    
  //  void locate(const svector<2>&, gelement&, svector<2>&) const;

  class geiterator : public std::vector<intArray<3> >::iterator
  {
  public:
    geiterator(const std::vector<intArray<3> >::iterator &it, triangleMesh *g) 
      : std::vector<intArray<3> >::iterator(it), gm(g) {}

    gelement operator*();
    
  private:
    triangleMesh * gm;
  };

  geiterator gebegin();
  geiterator geend();

  class bgeiterator : public std::vector<intArray<3> >::iterator
  {
  public:
    bgeiterator(const std::vector<intArray<3> >::iterator &it, triangleMesh *g)
      : std::vector<intArray<3> >::iterator(it), gm(g) {}

    bgelement operator*();
    
  private:
    triangleMesh * gm;
  };
  
  bgeiterator bgebegin();
  bgeiterator bgeend();
  
private:
  int nelem;
  int nodes;
  int nbelem;

  std::vector<svector<2> > vertices;       // vertex coordinates

  std::vector<intArray<3> > segments;      // boundary edges + colors

  std::vector<intArray<3> > triangles;     // triangles

  void input(std::string meshFile);         // auxillary function to read data

  friend class triangleVmesh;
};

triangleMesh::triangleMesh(std::string meshFile) 
{
  std::cout << "Initializing triangleMesh(" << meshFile << "):" 
	    << " (reads .node, .poly, and .ele files)" << std::endl;

  input(meshFile);

  // info();
}

void triangleMesh::info()
{
  std::cout << "Number of vertices read in = ..................... " << nodes
	    << std::endl;
  std::cout << "Number of boundary segments = .................... " << nbelem 
	    << std::endl;
  std::cout << "Number of triangles = ............................ " << nelem 
	    << std::endl;

  return;
}

gelement triangleMesh::geiterator::operator*() 
{
  gelement ee;
  int j, *tt = std::vector<intArray<3> >::iterator::operator*();
  
  for(j = 0; j < 3; j++) 
    {
      ee.xx[j] = gm->vertices[tt[j]];
      ee.nodes[j] = tt[j];
    }
  
  ee.master = &Triangle;
  
  return ee;
}

triangleMesh::geiterator triangleMesh::gebegin() 
{
  return geiterator(triangles.begin(), this);
}

triangleMesh::geiterator triangleMesh::geend() 
{
  return geiterator(triangles.end(), this);
}

bgelement triangleMesh::bgeiterator::operator*() 
{
  bgelement be;
  int j, *tt = std::vector<intArray<3> >::iterator::operator*();

  be.type = tt[2];

  for(j = 0; j < 2; j++) 
  {
      be.xx[j] = gm->vertices[tt[j]];
      
      be.nodes[j] = tt[j];
  }  

  be.master = &Interval;

  return be;
}

triangleMesh::bgeiterator triangleMesh::bgebegin() 
{
  return bgeiterator(segments.begin(), this);
}

triangleMesh::bgeiterator triangleMesh::bgeend() 
{
  return bgeiterator(segments.end(), this);
}

void triangleMesh::print()
{
  int i;

  std::ofstream dump("outputGmesh");

  dump.setf(std::ios::fixed, std::ios::floatfield);

  dump << "Number of gelements = " << nelem << std::endl;
  dump << "Number of nodes     = " << nodes << std::endl;

  i = 0;

 for (geiterator geit = gebegin(); 
      geit != geend(); geit++ )
   {
     gelement ge = *geit;

     dump << "gElement " << i++ << ": nodes[] = ";

     for(int j = 0; j < 3; j++) dump << ge.nodes[j] << "  ";

      dump << "\n\t (" 
	   << ge.xx[0] << "," << ge.xx[1] << "," << ge.xx[2] 
	   << ")\n"; 
    }

  dump << "\nNumber of boundary gelements = " << nbelem << std::endl;

  i = 0;

  for (bgeiterator beit = bgebegin(); 
      beit != bgeend(); beit++ )
   {
     bgelement be = *beit;

     dump << "Boundary gElement " << i++ << ": nodes[] = ";

     for(int j = 0; j < 2; j++) dump << be.nodes[j] << "  ";

      dump << "\n\t (" 
	  << be.xx[0] << "," 
	  << be.xx[1] << "), type = " << be.type << std::endl;
    }

  dump.close();
}

void triangleMesh::input(std::string meshFile)
{
  int nAttr, nBmkr, color, dim, index, i0,i1, i2;
  double xx,yy;

  std::ifstream in((meshFile + ".node").c_str());

  if( !in.is_open() ) 
    {
      std::cerr << "can't open file:" << meshFile + ".node" << std::endl;

      throw("Can't open file");
    }

  // Read in the vertices

  in >> nodes >> dim >> nAttr >> nBmkr;

  assert(dim == 2);

  assert(nodes > 0);

  vertices.resize(nodes);

  for(int i = 0; i < nodes; i++)
    {
      in.ignore(256,'\n');

      in >> index;

      if(index != i) 
	std::cerr << "strToMesh(): Warning: data in .nodes may be bad\n";

      in >> xx >> yy;

      vertices[i] = svector<2>(xx,yy);

      // ignore the rest
    }

  in.close();

  // Read in segments from the .poly file

  in.open((meshFile + ".poly").c_str());

  if( !in.is_open() ) 
    {
      std::cerr << "can't open file:" << meshFile + ".poly" << std::endl;

      throw("Can't open file");
    }

  in >> i0 >> dim >> nAttr >> nBmkr;

  assert(dim == 2);

  assert(i0 == 0);    // i0 = 0 means vertices in .node file
    
  in.ignore(256,'\n');
  
  in >> nbelem >> nBmkr;

  color = 0;      // default color if nBmkr = 0 (Dirichlet bc)

  segments.resize(nbelem);

  for(int i = 0; i < nbelem; i++)
    {
      in.ignore(256,'\n');

      in >> index;

      if(index != i) 
	std::cerr << "Warning: data in .poly file may be bad" << std::endl;

      in >> i0 >> i1;

      assert((-1 < i0) && (i0 < nodes));
      assert((-1 < i1) && (i1 < nodes));

      segments[i][0] = i0;
      segments[i][1] = i1;

      if(nBmkr > 0) in >> color;

      segments[i][2] = color-1;   // triangle messes up zero labels
    }

  in.close();

  // Read in the triangles

  in.open( (meshFile + ".ele").c_str() );

  if( !in.is_open() ) 
    {
      std::cerr << "can't open file:" << meshFile + ".ele" << std::endl;

      throw("Can't open file");
    }

  in >> nelem >> i0 >> nAttr;

  int oriented = 0;
  svector<2> v1, v2;

  triangles.resize(nelem);

  for(int i = 0; i < nelem; i++)
    {
      in.ignore(256,'\n');

      in >> index;

      if(index != i) 
	std::cerr << "Warning: data in .ele file may be bad" << std::endl;

      in >> i0 >> i1 >> i2;

      assert((-1 < i0) && (i0 < nodes));
      assert((-1 < i1) && (i1 < nodes));
      assert((-1 < i2) && (i2 < nodes));

      triangles[i][0] = i0;
      triangles[i][1] = i1;
      triangles[i][2] = i2;

      // check orientation

      v1 = vertices[i1] - vertices[i0];
      v2 = vertices[i2] - vertices[i0];

      double cross = v1[0]*v2[1] - v1[1]*v2[0];

      if(cross < 0.0)   // correct orientation
	{
	oriented ++;

	triangles[i][1] = i2;
	triangles[i][2] = i1;
	}

      // ignore the rest
    }

  in.close();

  if(oriented > 0)
    std::cerr << "Warning: triangleMesh::input(str): " << oriented 
	      << " triangles were not posativly oreiented; fixup attempted\n";
    
  checkBdyOrientation();

  return;
}

bool triangleMesh::checkBdyOrientation()
{
  // construct the topological boundary
  // and verify segments on it are consistently oriented
  // assumes that the triangles are correctly oriented

  typedef std::pair<int, int> intPair;

  set<intPair> edges;

  assert(edges.empty());

  for(int i = 0; i < nelem; i++)  // construct topolocial boundary
    {
    intPair eplus, eminus;

    for(int j = 0; j < 3; j++)
      {
      eplus  = intPair(triangles[i][j], triangles[i][(j+1)%3]);
      eminus = intPair(triangles[i][(j+1)%3], triangles[i][j]);

      set<intPair>::iterator loc = edges.find(eminus);
      
      if(loc != edges.end()) edges.erase(loc);
      else                   edges.insert(eplus);
      }
    }

  bool test = true;
  int nbdy = 0;

  // check orientation of segments on boundary

  for(int i = 0; i < nbelem; i++)
    {
    intPair eplus, eminus;

    eplus  = intPair(segments[i][0], segments[i][1]);
    eminus = intPair(segments[i][1], segments[i][0]);

    set<intPair>::iterator loc = edges.find(eplus);
      
    if(loc == edges.end())  // not found
      {
      loc = edges.find(eminus);

      if(loc != edges.end()) // incorrectly oriented
	{
	test = false;
	nbdy ++;

	segments[i][0] = eminus.first;
	segments[i][1] = eminus.second;

	edges.erase(loc);
	}
      }
    else     // found with correct orientation
      {
      nbdy++;

      edges.erase(loc);
      }
    }

  assert((int) edges.size() == nbelem-nbdy);

  if(!test)
    std::cerr << "triangleMesh::checkBdyOrientation(): "
	      << "boundary orientation test failed; fixup attempted\n";

  if(nbdy < nbelem)
    std::cerr << "triangleMesh::checkBdyOrientation(): " << nbdy
	      << " segments are on the topological boundary\n";


  return(test);
}

