#ifndef __VCELL_C__
#define __VCELL_C__

#include<vector>
#include"masterCell.C"
#include"../LinearAlgebra/svector.C"

using namespace std;

template<const int Ndim>
class VoronoiCell
{
public:
  VoronoiCell()                                    // Empty constructor
    : type(0), nNbr(0), neighbors(0), 
      xNeighbors(0), hEdges(0), aFaces(0)
  {return;}

  VoronoiCell(int tt, int nnbr)          // Constructor
    : type(tt), nNbr(nnbr), neighbors(nnbr), 
      xNeighbors(nnbr), hEdges(nnbr),  aFaces(nnbr)
  {return;}

  int type;                   // Index to specify material properties
  int nNbr;                   // Number of neighbors

  int node;                   // This cells node number
  svector<Ndim> xx;           // Coordinates of cell center
  double volume;              // Area/volume of this cell

  vector<int> neighbors;             // Node numbers of neighbors
  vector<svector<Ndim> > xNeighbors; // Coordinates of neighbors
  vector<double> hEdges;             // Lenghts of primal edges
  vector<double> aFaces;             // Length/areas of dual edges/faces  
};


template<const int Ndim>
class BoundaryVcell
{
public:
  BoundaryVcell()                                    // Empty constructor
    : type(0), nNbr(0), neighbors(0), 
      xNeighbors(0), hEdges(0),  aFaces(0)
  {return;}

  BoundaryVcell(int tt, int nnbr)          // Constructor
    : type(tt), nNbr(nnbr), neighbors(nnbr), 
      xNeighbors(nnbr), hEdges(nnbr),  aFaces(nnbr)
  {return;}

  int type;                   // Index to specify material properties
  int nNbr;                   // Number of neighbors

  int node;                   // This cells node number
  svector<Ndim> xx;           // Coordinates of cell center
  double area;                // Area/volume of this cell

  vector<int> neighbors;             // Node numbers of neighbors
  vector<svector<Ndim> > xNeighbors; // Coordinates of neighbors
  vector<double> hEdges;             // Lenghts of primal edges
  vector<double> aFaces;             // Primal boundary Lengths/areas
  vector<svector<2> > normals;       // Normals to boundary

  // Note: 1) In two dimensions there are always 2 neighbors.
  //       2) (a) In two dimeensions aFaces[.] = 0.5 hEdges[.]
  //          (b) In three dimensions its the area of the "kite"
  //         In each case, area is sum(aFaces[.])
};

#endif                          // !defined(__VCELL_C__)
