#include"twoPhase.h"

// two phase flow problem, implicit Euler approximation
// two phase flow problem, implicit Euler approximation
// two phase flow problem, implicit Euler approximation

double Normt[3]  = {0,0,0};   // L^2[0,T;L2] norm of p, ph, p-ph
double Dnormt[3] = {0,0,0};   // L^2[0,T;H1] norm of p, ph. p-ph

void (*exact)(double t, dvector xx, 
	      double pp[Nphase], dvector dp[Nphase]);

double (*gammast)(double sm, double pp[], double dg[], double d2g[][Nphase]);

double (*coeff)(const double t, const dvector xx, 
		double p0[Nphase], double p1[Nphase], 
		double &sm, double ss[Nphase],
		double dq[Nphase], double d2q[Nphase][Nphase], 
		dvector bb[Nphase], dmatrix kk[Nphase]);

ofstream dump; 

int main(int argc, char *argv[])
{
  int eg = 0, nitter0 = 0, nitter1 = 0, nstep = 8;
  int itter0, itter1, n, nprint, nplot;

  feMV mvclass(argc, &argv);

 if(argc > 1)
   {
   istringstream ist(argv[1]);

   if( !(ist >> eg) )
     {
     cerr << "Invalid value for eg:" << argv[1] << ", terminating" << endl;
     throw("Invalid input");
     }

   if((eg < 0) || (eg >= EgMax)) 
     {
     cerr << "Invalid value for eg:" << eg << ", terminating" << endl;
     throw("Invalid input argument");
     }
   }

 if(argc > 2)
   {
   istringstream ist(argv[2]);

   if( !(ist >> nstep) )
     {
     cerr << "Invalid value for nstep:" << argv[2]
	  << ", terminating" << endl;
     throw("Invalid input");
     }

   if(nstep <= 0) 
     {
     cerr << "Invalid value for nstep:" << nstep << ", terminating" << endl;
     throw("Invalid input argument");
     }
   }

 cout << "Two phase porous flow with Implicit Euler time stepping" << endl;
 cout << "Example " << eg << ", " << Examples[eg].desc << endl;
 cout << "Solution of " << Ndim << "d equation" << endl;
 cout << "Number of time steps (nstep) ................. = " << nstep<< endl;

 // Set the data for the example

 (Examples[eg].initialize)();
 exact   = Examples[eg].exact;
 gammast = Examples[eg].gammast;
 coeff   = Examples[eg].coeff;

 Tau = T / nstep;
 
 cout << "Estimated bound for (gamma*)'' (Beta) .......... " << Beta << endl;
 cout << "Final time (T) = ............................... " << T << endl;
 cout << "Time step size (Tau ) = ........................ " << Tau << endl;

 // Construct the mesh and dof

 string ss = "";
 if(argc > 2) ss = argv[2];  // default nstep = nex for quad and hex meshes
 if(argc > 3) ss = argv[3];    
 
 gmMesh mm(ss);

 mm.info();

 cout << "Number of elements = ........................... " 
      << mm.nElem() << endl;
 cout << "Number of space dof per step = ................. " 
      << 2*mm.nNode() << endl;

 // Allocate the global matrix and vectors

 feMV::matrix A(2*mm.nNode(), 2*mm.nNode());
 feMV::vector p0(2*mm.nNode());     // solution at t0
 feMV::vector pp(2*mm.nNode());     // solution at t0 + Tau

 // Specify the initial data

 double t = 0.0;

 initializeP(pp, mm);         // Interpolation intial values

 cout << endl;

  cout << " Time              L2 norms     "
       << "                   H1 semi-norms      " << endl;  

  cout << "          exact    approx      error"
       << "         exact    approx      error" << endl;

  norm(t, pp, mm);                  // Interpolation error
 
 // Set up time stepping

 nprint = (int) (nstep / 8.0 + 1.0e-08);
 if(nprint == 0) nprint = 1;

 nplot = (int) (nstep / 64.0 + 1.0e-08);
 if(nplot == 0) nplot = 1;

 nplot = 1;

 dump.open("dump");

 VtkFile vtkFile("Output/twoPhase", mm);  
 vtkFile.setTitle("Two Phase Code");

 vtkSaturations(t, vtkFile, pp, mm);

 Times.initializeC = clock() - Times.startC;
 Times.initializeW = difftime(time(NULL), Times.startW);

 for(n = 0; n < nstep; n++)
   {
   clock_t tstartC = clock();
   time_t  tstartW = time(NULL);

   itter0 = stepNesterov(t, pp, p0, A, mm);
   
   Times.nesterovC += clock() - tstartC;
   Times.nesterovW += difftime(time(NULL), tstartW);

   itter1 = 0;

   if(itter0 < 0) 
     {
     itter0 = -itter0;

     tstartC = clock();
     tstartW = time(NULL);

     itter1 += stepNewton(t, pp, p0, A, mm);

     Times.newtonC += clock() - tstartC;
     Times.newtonW += difftime(time(NULL), tstartW);
     }

     nitter0 += itter0;
     nitter1 += itter1;

   dump << "Time step n = " << n << " took " 
	<< itter0 << " Nesterov itterations and "
	<< itter1 << " Newton itterations" << endl;

   normt(t, pp, p0, mm);   // accumulate the L^2[0,T; .] norms
   
   t += Tau;
   
   if((n+1) % nprint == 0) norm(t, pp, mm);

   if((n+1) % nplot == 0)
     vtkSaturations(t, vtkFile, pp, mm);
   }

 cout << endl;

 normtPrn(t);

 cout << endl;
 cout << "Total number of Nesterov itterations = " << nitter0 << endl;
 cout << "Total number of Newton   itterations = " << nitter1 << endl;

 Times.print();

 dump.close();

 return(0);
}

int stepNewton(double t0, feMV::vector &p1,
	 feMV::vector &p0, feMV::matrix &A, gmMesh &mm)
{
  int Itmax = 200, itter, nvar = 2*mm.nNode();

  double epsilon = 1.0e-8, beta = 1;
  double delta = 0, psi, psiOld = 0, dpsiNorm = 0;
  
  feMV::vector dpsi(nvar);      // gradient of Psi
  feMV::vector pOld(nvar);      // prior guess
  feMV::vector delp(nvar);      // Newton update

  // p0 = p1;               // solution at t0 .... set in stepNesterov()

  for(itter = 0; itter < Itmax; itter++)
    {
    psi = dPsi(t0, p1, p0, dpsi, A, true, mm, false); 

    dpsiNorm = dpsi.normInfinity();

    if(isnan(dpsiNorm)) break;

    if((itter == 0) || (dpsiNorm < epsilon))       // print first and last
      dump << "stepNewton():   Itteration " << itter 
	   << ", |dpsi| = " << dpsiNorm 
	   << ", psi = " << psi 
	   << ", delta = " << delta 
	   << endl;
    
    if(dpsiNorm < epsilon) return(itter+1);

    if(itter == 0) psiOld = psi + 1;

    if(psi > psiOld)  // reject prior step
      {
	beta *= 0.5;
	
	p1 = pOld - beta * delp;
      }
    else               // accept prior step
      {
	pOld = p1;
	psiOld = psi;

	clock_t tstartC = clock();
	time_t  tstartW = time(NULL);

	delp = A.solve(dpsi);

	Times.solveC += clock() - tstartC;
	Times.solveW += difftime(time(NULL), tstartW);
	
	delta = delp.normInfinity();
	
	p1 = p1 - beta * delp;         // Newton update
	
	beta = fmin(1.0, 1.25 * beta); 
      }
    }

  dump << "stepNewton(): Itteration " << itter 
       << ", |dpsi| = " << dpsiNorm 
       << ", psi = " << psi 
       << ", delta = " << delta 
       << endl;
  
  cout << "stepNewton(): didn't converge in " << itter
       << " itterations, beta = " << beta << endl;
  cout << "Time t0 of final itteration = " << t0 
       << ", psi  = " << psi <<  endl;
  cout << "Delta  for final itteration = " << delta 
       << ", dpsi = " << dpsiNorm <<  endl;
  
  throw("Didn't converge");
  
  // return(-itter);
}

int stepNesterov(double t0, feMV::vector &p1,
	 feMV::vector &p0, feMV::matrix &A, gmMesh &mm)
{
  int Itmax = 50, itter = 1, nvar = mm.nNode();

  double psi, lambda = 1.0, gamma, epsilon = 1.0e-8;
  
  feMV::vector dpsi(2*mm.nNode());      // gradient of Psi
  feMV::vector gg(2*mm.nNode());
  feMV::vector y0(2*mm.nNode());
  feMV::vector y1(2*mm.nNode());

  p0 = p1;            // p0 = gradPsi(t0)
  y0 = p1;            // p1 = initial guess (xx in notes)

  psi = dPsi(t0, p1, p0, dpsi, A, true, mm, true);  // compute preconditioner
  
  double dpsiNorm = dpsi.normInfinity();

  for(itter = 0; itter < Itmax; itter++)
    {
    if(isnan(dpsiNorm)) break;

    if((itter == 0) || (dpsiNorm < epsilon)) // print first and last itteration
      dump << "stepNesterov():   Itteration " << itter 
	   << ", |dpsi| = " << dpsiNorm 
	   << ", psi = " << psi 
	   << endl;
    
    if(dpsiNorm < epsilon) return(itter+1);

    if(itter == 0) A.solve(gg, dpsi);      // will compute factorization
    else           A.reSolve(gg, dpsi);

    y1 = p1 - gg;

    gamma = 1 - lambda;
    lambda = 0.5 * (1 + sqrt(1+4*lambda*lambda));
    gamma /= lambda;

    p1 = (1-gamma) * y1 + gamma * y0;
    y0 = y1;

    psi = dPsi(t0, p1, p0, dpsi, A, false, mm, true); // don't compute d2psi

    dpsiNorm = dpsi.normInfinity();
    }

  dump << "stepNesterov(): Itteration " << itter 
       << ", |dpsi| = " << dpsiNorm 
       << ", psi = " << psi 
       << endl;

  return(-itter);
}
 
void  initializeP(feMV::vector &pp, gmMesh &mm)
{
  int i, pi;
  double t = 0.0, pe[Nphase];
  dvector dpe[Nphase];

  // Specify the initial values 

 for (gmMesh::viterator cit = mm.vbegin(); cit != mm.vend(); cit++ )
    {
    vcell cc = *cit;

    exact(t, cc.xx, pe, dpe);
	
    for(pi = 0; pi < Nphase; pi++) pp[cc.node + pi*mm.nNode()] = pe[pi];
    }

  return;  
}

void norm(double t, feMV::vector &pp, gmMesh &mm)
{
  int i, pi;
  double pe[Nphase], pej[Nphase], ph[Nphase];
  double norm[] = {0,0,0}, dnorm[] = {0,0,0};

  dvector dpe[Nphase];

 for (gmMesh::viterator vit = mm.vbegin(); vit != mm.vend(); vit++ )
    {
    vcell cc = *vit;

    exact(t, cc.xx, pe, dpe);
	
    for(pi = 0; pi < Nphase; pi++)
      {
      ph[pi] = pp[cc.node + pi*mm.nNode()];
	
      norm[0] += pe[pi]*pe[pi] * cc.volume;
      norm[1] += ph[pi]*ph[pi] * cc.volume;
      norm[2] += (pe[pi]-ph[pi])*(pe[pi]-ph[pi]) * cc.volume;
      }

    for(int j = 0; j < cc.nNbr; j++)
      {
      if(cc.neighbors[j] < cc.node) continue;  // dont double count edges

      double aij = cc.aFaces[j] / cc.hEdges[j];

      exact(t, cc.xNeighbors[j], pej, dpe);
      
      for(pi = 0; pi < Nphase; pi++)
	{
	double tempe = pe[pi] - pej[pi];
	double temph = ph[pi] - pp[cc.neighbors[j] + pi*mm.nNode()]; 
	
	dnorm[0] += aij * tempe*tempe;
	dnorm[1] += aij * temph*temph;
	dnorm[2] += aij * (tempe-temph)*(tempe-temph);
	} 
      }
    }

  for(i = 0; i < 3; i++)
    {
      norm[i]  = sqrt( norm[i]);
      dnorm[i] = sqrt(dnorm[i]);
    }

  cout.setf(ios::fixed,ios::floatfield);
  cout.width(7);
  cout.precision(4);    
  cout << t;

  cout.width(10);
  cout.precision(6);
  cout << norm[0];
    
  cout.width(10);
  cout.precision(6);
  cout <<  norm[1];
    
  cout.setf(ios::scientific,ios::floatfield);
  cout.width(14);
  cout.precision(6);
    
  cout << norm[2] << " " ;
    
  cout.setf(ios::fixed,ios::floatfield);
  cout.width(10);
  cout.precision(6);
  cout <<  dnorm[0];

  cout.width(10);
  cout.precision(6);
  cout <<  dnorm[1];
    
  cout.setf(ios::scientific,ios::floatfield);
  cout.width(14);
  cout.precision(6);  
  cout <<  dnorm[2] << endl;

  return;

 error:
 cerr << "Negative Jacobian in norm()" << endl;
 throw("Negative Jacobian");
}

void normt(double t0, feMV::vector &pp, feMV::vector &pold, gmMesh &mm)
{
  // Accumulate the L^2[0,T,L^2] and L^2[0,T,H^1] norms
  // using mid point rule in time.

  int i, pi;
  double pe[Nphase], pej[Nphase], ph[Nphase], weight;

  dvector dpe[Nphase];

  clock_t tstartC = clock();
  time_t  tstartW = time(NULL);

 for (gmMesh::viterator vit = mm.vbegin(); vit != mm.vend(); vit++ )
    {
    vcell cc = *vit;

    exact(t0+Tau/2, cc.xx, pe, dpe);
	
    weight = Tau * cc.volume;

    for(pi = 0; pi < Nphase; pi++)
      {
      ph[pi] = 0.5 * (pp[cc.node + pi*mm.nNode()] +
		      pold[cc.node + pi*mm.nNode()]);
      
      Normt[0] += pe[pi]*pe[pi] * weight;
      Normt[1] += ph[pi]*ph[pi] * weight;
      Normt[2] += (pe[pi]-ph[pi])*(pe[pi]-ph[pi]) * weight;
      }

    for(int j = 0; j < cc.nNbr; j++)
      {
      if(cc.neighbors[j] < cc.node) continue;  // dont double count edges

      weight = Tau * cc.aFaces[j] / cc.hEdges[j];

      exact(t0+Tau/2, cc.xNeighbors[j], pej, dpe);
	
      for(pi = 0; pi < Nphase; pi++)
	{
	double tempe = pe[pi] - pej[pi];
	double temph = ph[pi] - pp[cc.neighbors[j] + pi*mm.nNode()]; 
	
	Dnormt[0] += weight * tempe*tempe;
	Dnormt[1] += weight * temph*temph;
	Dnormt[2] += weight * (tempe-temph)*(tempe-temph);
	}
      }
    }

  Times.normtC += clock() - tstartC;
  Times.normtW += difftime(time(NULL), tstartW);

 return;

 error:
 cerr << "Negative Jacobian in normt()" << endl;
 throw("Negative Jacobian");
}

void normtPrn(double t)
{
  for(int i = 0; i < 3; i++)
    {
      Normt[i]  = sqrt( Normt[i]);
      Dnormt[i] = sqrt(Dnormt[i]);
    }

  cout << "              L^2[0,T;L^2] norms"
       << "               L^2[0,T;H^1] semi-norms" << endl;  

  cout << "          exact    approx      error"
       << "         exact    approx      error" << endl;

  cout.setf(ios::fixed,ios::floatfield);
  cout.width(7);
  cout.precision(4);    
  cout << t;

  cout.width(10);
  cout.precision(6);
  cout << Normt[0];
    
  cout.width(10);
  cout.precision(6);
  cout <<  Normt[1];
    
  cout.setf(ios::scientific,ios::floatfield);
  cout.width(14);
  cout.precision(6);
    
  cout << Normt[2] << " " ;
    
  cout.setf(ios::fixed,ios::floatfield);
  cout.width(10);
  cout.precision(6);
  cout <<  Dnormt[0];

  cout.width(10);
  cout.precision(6);
  cout <<  Dnormt[1];
    
  cout.setf(ios::scientific,ios::floatfield);
  cout.width(14);
  cout.precision(6);  
  cout <<  Dnormt[2] << endl;

return;
}

void dirichlet(double t, dvector xx, double pp[Nphase])
{
  dvector dp[Nphase];

  exact(t,xx, pp, dp);

  return;
}

void neumann(double t0, dvector xx, dvector normal, double gg[Nphase])
{
  double p0[Nphase], p1[Nphase], s0[Nphase], sm, smj;
  double dq[Nphase], d2q[Nphase][Nphase];
  dvector bb[Nphase], dp[Nphase];
  dmatrix kk[Nphase];

  exact(t0,xx, p0,dp);      // coefficients K(grad(p) - b).n at old time

  exact(t0+Tau,xx, p1,dp);   // grad(p) at new time

  coeff(t0,xx,p0,p1, sm,s0, dq,d2q, bb,kk); 
  
  for(int pi = 0; pi < Nphase; pi++) 
    gg[pi] = normal.dot(kk[pi]*(dp[pi]-bb[pi])); 

  return;
}

double dPsi(double t0, feMV::vector &p1, feMV::vector &p0, 
	    feMV::vector &dpsi, feMV::matrix &A, bool hessian, 
	    gmMesh &mm, bool nesterov)
{
  bool flag[Nphase];

  int ii,jj, pi,pj, npres = mm.nNode();

  double epsilon = 1.0e-8,  psi = 0.0;
  double p0i[Nphase], p1i[Nphase], p0j[Nphase], p1j[Nphase];
  double gst, qq, sm, smj, temp, tempk, s0[Nphase];
  double dq[Nphase], d2q[Nphase][Nphase];
  double dgst[Nphase], d2gst[Nphase][Nphase];
  dvector bi[Nphase], bj[Nphase];
  dmatrix ki[Nphase], kj[Nphase];

  vector< Eigen::Triplet<double> > Aentries;

  dpsi.setZero();

  for (gmMesh::viterator vit = mm.vbegin(); vit != mm.vend(); vit++)
    {
    vcell cc = *vit;

    for(pi = 0; pi < Nphase; pi++)
      {
      p0i[pi] = p0[cc.node + pi*npres];
      p1i[pi] = p1[cc.node + pi*npres];

      flag[pi] = true;
      }

    qq = coeff(t0,cc.xx, p0i,p1i, sm,s0, dq,d2q, bi,ki); // bi, ki, s0 at (n-1)

    gst = gammast(sm,p1i, dgst, d2gst);        // saturations at current time

    assert(Nphase == 2);      // Code for diagonal terms assumes Nphase = 2

    if(nesterov)             // use an upper bound for the Hessian
      {
      double beta = Beta * (1-sm)*sqrt(1-sm);

      // gst += 0.5 * epsilon * (p1i[0]*p1i[0] + p1i[1]*p1i[1]);

      // dgst[0] += epsilon * p1i[0];
      // dgst[1] += epsilon * p1i[1];

      d2gst[0][0] += beta + epsilon;
      d2gst[1][1] += beta + epsilon;
      d2gst[0][1] -= beta;
      d2gst[1][0] -= beta;
      }
    else 
      {
      d2gst[0][0] += epsilon;
      d2gst[1][1] += epsilon;
      }

    psi += (gst + Tau*qq) * cc.volume;

    for(pi = 0; pi < Nphase; pi++)  // Temporal/diagonal terms
      {
      ii = cc.node;

      psi -= s0[pi] * p1[ii+pi*npres] * cc.volume;

      dpsi[ii+pi*npres] +=  cc.volume * (dgst[pi] - s0[pi] + Tau*dq[pi]);

      if(hessian)
	for(pj = 0; pj < Nphase; pj++)
	  Aentries.push_back(Eigen::Triplet<double>
	       (ii+pi*npres, ii+pj*npres, 
		(d2gst[pi][pj] + Tau*d2q[pi][pj]) * cc.volume));
      }

    if((qq < epsilon) && (p1i[0] - p1i[1] > -0.5/sqrt(1-sm)))
      {
	if(ki[0].frobDot(ki[0]) > epsilon) flag[0] = false;
	if(ki[1].frobDot(ki[1]) > epsilon) flag[1] = false;
      }
    else
      flag[0] = flag[1] = false;          // Q() or gammast() not degenerate

    for(int j = 0; j < cc.nNbr; j++)      // Spatial terms
      {
      if(cc.neighbors[j] < cc.node) continue;  // dont double count edges

      double wij = Tau * cc.aFaces[j] / cc.hEdges[j];

      for(pi = 0; pi < Nphase; pi++)
	{
	  p0j[pi] = p0[cc.neighbors[j] + pi*npres];
	  p1j[pi] = p1[cc.neighbors[j] + pi*npres];
	}

      coeff(t0,cc.xNeighbors[j], p0j,p1j, smj,s0, dq,d2q, bj,kj); 

      for(pi = 0; pi < Nphase; pi++)
	{
	ii = cc.node + pi*npres;
	jj = cc.neighbors[j] + pi*npres;

	if(kj[pi].frobDot(kj[pi]) > epsilon) flag[pi] = false;

	double kij 
	  = wij * 0.5*(ki[pi][0][0] + kj[pi][0][0]); // ki, kj diagonal

	temp = p1[ii] - p1[jj]
	  - 0.5*(bi[pi] + bj[pi]).dot(cc.xNeighbors[j]-cc.xx);

	tempk = kij * temp;

	psi += 0.5 * tempk * temp;

	dpsi[ii] += tempk;
	dpsi[jj] -= tempk;

	if(hessian)
	  {
	    Aentries.push_back(Eigen::Triplet<double>(ii,ii,  kij));
	    Aentries.push_back(Eigen::Triplet<double>(ii,jj, -kij));
	    Aentries.push_back(Eigen::Triplet<double>(jj,jj,  kij));
	    Aentries.push_back(Eigen::Triplet<double>(jj,ii, -kij));
	  }
	}
      }

    if(flag[0] | flag[1])    // Degenerate region
      {
      ii = cc.node;
      jj = ii + npres;
      
      temp = p1[ii]-p1[jj] - 0.5/sqrt(1-sm);

      psi += 0.5 * temp*temp * cc.volume;

      temp *= cc.volume;

      dpsi[ii] += temp;
      dpsi[jj] -= temp;
      
      if(hessian)
	{
	  Aentries.push_back(Eigen::Triplet<double>(ii,ii,  cc.volume));
	  Aentries.push_back(Eigen::Triplet<double>(ii,jj, -cc.volume));
	  Aentries.push_back(Eigen::Triplet<double>(jj,ii, -cc.volume));
	  Aentries.push_back(Eigen::Triplet<double>(jj,jj,  cc.volume));
	}
      }
    }

  if(exact == exact0)
    {
    double big = 1.0e16;

    for (gmMesh::bviterator bvit = mm.bvbegin(); bvit != mm.bvend(); bvit++ )
      {
      bvcell cc = *bvit;     

      dirichlet(t0+Tau,cc.xx, p1i);

      for(pi = 0; pi < Nphase; pi++)
	{
	ii = cc.node + pi*npres;

	p1[ii] = p1i[pi];

	dpsi[ii]  = 0         ;
	
	if(hessian) Aentries.push_back(Eigen::Triplet<double>(ii,ii, big));
	}
      }
    }

  if(hessian)
    {
      A.setZero();

      A.setFromTriplets(Aentries.begin(), Aentries.end());
    }

  return(psi);
}


void vtkSaturations(double t, VtkFile &vtkFile, feMV::vector &pp, gmMesh &mm)
{
  int i, pi;
  double pe[Nphase], se[Nphase], d2gst[Nphase][Nphase];

  // compute the saturations

  feMV::vector sat(pp.size());     // saturations

  for (gmMesh::viterator vit = mm.vbegin(); vit != mm.vend(); vit++)
    {
    vcell cc = *vit;

    for(pi = 0; pi < Nphase; pi++) pe[pi] = pp[cc.node + pi*mm.nNode()];

    double sm = porosity1(cc.xx);

    gammast(sm, pe, se, d2gst);

    for(pi = 0; pi < Nphase; pi++) 
      sat[cc.node + pi*mm.nNode()] = se[pi]/(1-sm + 1.0e-6);

    if(Ndim == 2) sat[cc.node + mm.nNode()] = se[0];  // Actual saturation
    }

  vtkFile.addScalarName("RelativeSaturation");    
  vtkFile.addScalarName("Saturation");    
  
  // vtkFile.write(t, pp);
  
  vtkFile.write(t, sat);

  return;  
}
