// Sample program to solve Poisson's equation using control volumes
// Sample program to solve Poisson's equation using control volumes
// Sample program to solve Poisson's equation using control volumes

#include<cstdio>
#include<iostream>
#include<iomanip>
#include<cmath>
#include<cassert>
#include<ctime>
#include<vector>

#include"../LinearAlgebra/Eigen/Eigen/Sparse"
#include"../LinearAlgebra/Eigen/Eigen/SparseLU"

using namespace std;

// options specified next -- (un)comment options as required
// options specified next -- (un)comment options as required
// options specified next -- (un)comment options as required

// The DeBug flag enables some (possibly slow) checks such
// as array bound checking.

#define DeBug

// Two or Three dimensions, basis functions

#define NDIM  2          // spatial dimension

// Mesh type

#define TPLCMESH
// #define PLCMESH

// end option specification -- code starts
// end option specification -- code starts
// end option specification -- code starts

class TimeClass
{
 public:
  clock_t startC, solveC, initializeC, assembleC, normtC, newtonC, nesterovC;
  time_t  startW, solveW, initializeW, assembleW, normtW, newtonW, nesterovW;

  TimeClass() 
    {
      startC = clock();           // CPU  time
      startW = time(NULL);        // Wall time

      solveC = initializeC = assembleC = normtC = newtonC = nesterovC = 0;
      solveW = initializeW = assembleW = normtW = newtonW = nesterovW = 0;
    }

  void print()
  {
    cout << endl;
    cout << "CPU  time to initialize problem = .... " 
	 << initializeC /  ((double) CLOCKS_PER_SEC)
	 << ", Wall time = " << initializeW  << endl;

    cout << "CPU  time to assemble matrices ....... " 
	 << assembleC /  ((double) CLOCKS_PER_SEC)
	 << ", Wall time = " << assembleW  << endl;

    cout << "CPU  time in normt() = ............... " 
	 << normtC /  ((double) CLOCKS_PER_SEC) 
	 << ", Wall time = " << normtW  << endl;

    cout << "CPU  time for linear solver .......... " 
	 << solveC /  ((double) CLOCKS_PER_SEC)
	 << ", Wall time = " << solveW  << endl;

    cout << "CPU  time for Newton itterates ....... " 
	 << newtonC /  ((double) CLOCKS_PER_SEC)
	 << ", Wall time = " << newtonW  << endl;

    cout << "CPU  time for Nesterov itterates...... " 
	 << nesterovC /  ((double) CLOCKS_PER_SEC)
	 << ", Wall time = " << nesterovW  << endl;

    cout << "Total CPU  time = .................... "
	 << (clock() - startC) /  ((double) CLOCKS_PER_SEC)
	 << ", Wall time = " << difftime(time(NULL), startW)  << endl;
  }
};

TimeClass Times;

// include small and large matrix classes and quadrature points

const int Ndim = NDIM;        // two or three dimensional code
const int Nphase = 2;         // two phase problem

double T, Tau;  // Final time and time step

const double Pi = 3.14159265358979323846;  

// Linear Algebra 

#include"../LinearAlgebra/smatrix.C"
#include"../LinearAlgebra/svector.C"

typedef smatrix<Ndim> dmatrix;
typedef svector<Ndim> dvector;

class feMV
{
public:
  class vector;
  class matrix;

  feMV() {}
  feMV(int &argc, char ***argv) {}

  ~feMV() {cout << "~feMV() Eigen matrix vector classes" << endl;}

  class matrix : public  Eigen::SparseMatrix<double> 
  {
  public:

    matrix(int m, int n) : Eigen::SparseMatrix<double>(m,n) {}

      Eigen::VectorXd solve(vector &F)
      {
	this->makeCompressed();

	solver.analyzePattern(*this); 

	solver.factorize(*this); 
 
 	if(solver.info() != Eigen::Success)  
 	  { 
 	    cerr << "solve(): LU decomposition failed, info() = " 
		 << solver.info() << endl ;  

 	    assert(false); 
 	  } 

	return( solver.solve(F) );
      }

      void solve(vector &U, vector &F) {U = solve(F); }

      Eigen::VectorXd reSolve(vector &F) { return( solver.solve(F) ); }

      void reSolve(vector &U, vector &F) { U = solver.solve(F); }

  private:
    Eigen::SparseLU<Eigen::SparseMatrix<double> > solver;
  };
    
  class vector : public Eigen::VectorXd 
  {
  public:
    vector(int m) : Eigen::VectorXd(m) {}
      
      vector(const Eigen::VectorXd &v) : Eigen::VectorXd(v) {}

	vector operator=(const Eigen::VectorXd &v) 
	  {	    
	    *this = vector(v);
	    return(*this);
	  }

	double normInfinity() {return lpNorm<Eigen::Infinity>();}

	double& operator[](int i) {return (*this)(i);}
  };
};

// Set up mesh

#include"../MeshGeneration/gelement.C"

#if NDIM == 1                
  const int Nchi = 2;          

  #include"../MeshGeneration/onedVmesh.C"
  typedef onedVmesh gmMesh;
#endif  // NDIM == 1

#if NDIM == 2                
    const int Nchi = 3;          

    #include"../MeshGeneration/triangleVmesh.C"
    typedef triangleVmesh gmMesh;

//    #include"../MeshGeneration/rectangleTriVmesh.C"
//    typedef rectangleTriVmesh gmMesh;

//    #include"../MeshGeneration/twodTPLCvMesh.C"
//    typedef twodTPLCvMesh gmMesh;
#endif    // NDIM == 2

#include"../Graphics/vtkFdmOutput.C"

#if NDIM == 3
    const int Nchi = 4;          

    #ifdef TPLCMESH                   // simple interface to read .tplc file  
    #include"../MeshGeneration/dir3vMeshTPLC.C"
    #elif defined PLCMESH             // dir3 constructs mesh from .plc file
    #include"../MeshGeneration/dir3vMeshPLC.C"
    #endif

    typedef dir3vMesh gmMesh;
#endif      // Ndim == 3

// Function Prototypes

int stepNewton(double t, feMV::vector &pp, 
	 feMV::vector &p0, feMV::matrix &A, gmMesh &mm);

int stepNesterov(double t, feMV::vector &pp, 
	 feMV::vector &p0, feMV::matrix &A, gmMesh &mm);

double dPsi(double t0, feMV::vector &p1, feMV::vector &p0, 
	    feMV::vector &dpsi, feMV::matrix &A, bool d2psi, 
	    gmMesh &mm, bool nesterov);

void norm(double t, feMV::vector &pp, gmMesh &mm);
void normt(double t0, feMV::vector &pp, feMV::vector &pold, gmMesh &mm);
void normtPrn(double t);

void dirichlet(double t, dvector xx, double pp[Nphase]);
void neumann(double t, dvector xx, dvector normal, double gg[Nphase]);

void initializeP(feMV::vector &pp, gmMesh &mm);

void vtkSaturations(double t, VtkFile &vtkFile, feMV::vector &pp, gmMesh &mm);

// STRUCTURE TO CONTAIN DATA ROUTINES FOR DIFFERENT EXAMPLES

typedef struct
  {
    /* Names:
       exact   = exact solution
       coeff   = saturations, wells, body force, permiability (t,x,p[]) 
       desc    = description
    */

    void (*initialize)(void);
    void (*exact)(const double t, const dvector xx, 
		  double pp[Nphase], dvector dp[Nphase]);
    double (*gammast)(const double sm, double pp[Nphase], 
		      double dg[Nphase], double d2gst[Nphase][Nphase]);
    double (*coeff)(const double t, const dvector xx, 
		    double p0[Nphase], double p1[Nphase], 		    
		    double &sm, double ss[Nphase],
		    double dq[Nphase], double d2q[Nphase][Nphase], 
		    dvector bb[Nphase], dmatrix kk[Nphase]);
    string desc;
  } example_data;

// sm = saturation of medium 
// s0[] = saturations of phases
// p0 = pressures at time (n-1), 
// p1[] = current guess of pressurs

// qq = coeff() is Q(p1) for the source terms
// dq[] = DQ(pi)
// d2q[][] = D^2(Q)(p1)

// bb[] = body forces
// kk[] = diffusion matrices (diagonal for control volume examples)

#if NDIM == 2

  const int EgMax = 3;

  #include"twoPhaseEg2d.C"     

  example_data Examples[] =   
    {
    {initialize0,exact0,gammast0,coeff0,"heat equation"},
    {initialize1,exact1,gammast1,coeff1,"one injection and one production"},
    {initialize2,exact2,gammast2,coeff2,"injection/production k0 != k1"}
    };
#endif

#if NDIM == 3

  const int EgMax = 1;

  #include"twoPhaseEg3d.C"     

  example_data Examples[] =   
    {
      {initialize0,exact0,gammast0,coeff0,"heat equation"}
    };
#endif

