double Beta = 1.0;   // Maximum value of (gamma*)''

void initialize0(void)   // heat equation
{
  assert(Nphase == 2);

  Beta = 0.0;

  T = 1.0;
}

void exact0(const double t, const dvector xx, double pp[2], dvector dp[2])
{
 double pix,piy, cos_pix,sin_piy, exp_t;

 pix = Pi * xx[0];
 piy = Pi * xx[1];

 cos_pix = cos(pix);
 sin_piy = sin(piy);
 exp_t   = exp(t); 

 pp[0] = pp[1] = cos_pix * sin_piy * exp_t;

 dp[0] = dp[1] = (Pi*exp_t) * dvector(-sin(pix)*sin_piy, cos_pix*cos(piy));

 return;
}

double gammast0(const double sm, double pp[2], double dg[2], double d2g[2][2])
{
  dg[0] = pp[0];
  dg[1] = pp[1];

  if(d2g == NULL) return( 0.5*(pp[0]*pp[0] + pp[1]*pp[1]) );

  d2g[0][0] = d2g[1][1] = 1.0;
  d2g[0][1] = d2g[1][0] = 0.0;

  return( 0.5*(pp[0]*pp[0] + pp[1]*pp[1]) );
}

double gammast0(const double sm, double pp[2], double dg[2])
{
  return( gammast0(sm,pp, dg, NULL) );
}

double coeff0(const double t, const dvector xx, 
	      double  p0[2], double p1[2],
	      double  &sm,   double ss[2],
	      double  dq[2], double d2q[2][2], 
	      dvector bb[2], dmatrix kk[2])
{
  sm = 0.5;

  gammast0(sm, p0, ss);

  double pe[2];
  dvector dpe[2];

  exact0(t,xx, pe,dpe);   // get the exact solution for RHS

  dq[0] = p1[0] - 2*(Pi*Pi + 1) * pe[0];
  dq[1] = p1[1] - 2*(Pi*Pi + 1) * pe[1];

  d2q[0][0] = d2q[1][1] = 1.0;
  d2q[1][0] = d2q[0][1] = 0.0;

  bb[0].zero(); bb[1].zero();
  kk[0] = kk[1] = dmatrix().identity();

  return( 0.5 * (dq[0]*dq[0] + dq[1]*dq[1]) ) ;
}

void initialize1(void)   // injection at (-1,-1), production at (1,1)
{
  assert(Nphase == 2);

  Beta = 8;

  T = 5.0;
}

double porosity1(dvector xx) 
{ return 0.7 + 0.1*sin(Pi*(2*xx[0]+xx[1])); }   // 1-(this) = porosity 

void exact1(const double t, const dvector xx, double pp[2], dvector dp[2])
{
  // Initially (almost) all phase 1 
  // p[0] - pp[1] << 0 so ss[0] almost 0

  pp[0] = -5.0;
  pp[1] =  5.0;
  
  dp[0] = dp[1] = dvector(0,0);

  return;
}

double gammast1(const double sm, double pp[2], double dg[2], double d2g[2][2])
{
  double pst = -0.5 / sqrt(1-sm), delp = pp[0]-pp[1], gst, dgst;

  double const eps = 1.0e-08;

  if(delp > pst) 
    {
      dgst = 1-sm;            
      gst = (1-sm)*delp; // + (eps/2)*(pp[0]*pp[0] + pp[1]*pp[1]);

      dg[0] = dgst; // + eps*pp[0]; 
      dg[1] = 1-sm - dgst; // + eps*pp[1];

      if(d2g == NULL) return( (1-sm)*pp[1] + gst );  

      d2g[0][0] = d2g[1][1] = eps;
      d2g[0][1] = d2g[1][0] = 0.0;

      return( (1-sm)*pp[1] + gst );   
    }
  else           
    {
      dgst = 1/(4*delp*delp); 
      gst = -1/(4*delp) - sqrt(1-sm); // + (eps/2)*(pp[0]*pp[0] + pp[1]*pp[1]);

      dg[0] = dgst; // + eps*pp[0]; 
      dg[1] = 1-sm - dgst; // + eps*pp[1];

      if(d2g == NULL) return( (1-sm)*pp[1] + gst );  

      d2g[0][0] = d2g[1][1] = -2.0*dgst/delp + eps;
      d2g[0][1] = d2g[1][0] =  2.0*dgst/delp;

      return( (1-sm)*pp[1] + gst );  
    }
}

double gammast1(const double sm, double pp[2], double dg[2])
{
  return( gammast1(sm,pp, dg, NULL) );
}

double coeff1(const double t, const dvector xx,
	      double p0[2], double p1[2], 
	      double  &sm,   double s0[2],
	      double  dq[2], double d2q[2][2], 
	      dvector bb[2], dmatrix kk[2])

{
  dvector c0 = xx - dvector(-1,-1);
  dvector c1 = xx - dvector( 1, 1);

  sm = porosity1(xx);

  gammast1(sm, p0, s0);

  bb[0].zero(); bb[1].zero();

  dvector cc = xx - dvector(0.25,-0.25);
  
  dmatrix km = 2.5 * fmax(exp(cc.dot(cc))-1.25, 0.0) * dmatrix().identity();

  double k0 = s0[0] * fmax(0.0, s0[0]-0.1*(1-sm));
  double k1 = s0[1] * fmax(0.0, s0[1]-0.1*(1-sm));

  double qq = 0.0;
  dq[0] = dq[1] = 0.0;
  d2q[0][0] = d2q[1][0] = d2q[0][1] = d2q[1][1] = 0.0;

  if(c0.dot(c0) < 1.0/64.0)   // injection well
    {
      dq[0] = p1[0] - 60;
      //      dq[1] = p1[1];
      
      qq = 0.5 * (dq[0]*dq[0] + dq[1]*dq[1]);

      d2q[0][0] = 1.0;
      //      d2q[1][1] = 1.0;

      //km.identity();
      //k0 += 0.25;   
      //k1 += 0.25;   
    }

  if(c1.dot(c1) < 1.0/64.0)   // production well
    {
      dq[0] = p1[0];
      dq[1] = p1[1];   // mean pressure zero on production well
      
      qq = 0.5 * (dq[0]*dq[0] + dq[1]*dq[1]);

      d2q[0][0] = d2q[1][1] = 1.0;

      //km.identity();
      //k0 += 0.25;   
      //k1 += 0.25;   
    }

  kk[0] = k0 * km;
  kk[1] = k1 * km;

  return(qq);
}

void initialize2(void)   // injection at (-1,-1), production at (1,1)
{
  assert(Nphase == 2);

  Beta = 8;

  T = 5.0;
}

double porosity2(dvector xx) 
{ return 0.7 + 0.1*sin(Pi*(2*xx[0]+xx[1])); }   // 1-(this) = porosity 

void exact2(const double t, const dvector xx, double pp[2], dvector dp[2])
{
  // Initially (almost) all phase 1 
  // p[0] - pp[1] << 0 so ss[0] almost 0

  pp[0] = -5.0;
  pp[1] =  5.0;
  
  dp[0] = dp[1] = dvector(0,0);

  return;
}

double gammast2(const double sm, double pp[2], double dg[2], double d2g[2][2])
{
  double pst = -0.5 / sqrt(1-sm), delp = pp[0]-pp[1], gst, dgst;

  double const eps = 1.0e-08;

  if(delp > pst) 
    {
      dgst = 1-sm;            
      gst = (1-sm)*delp; // + (eps/2)*(pp[0]*pp[0] + pp[1]*pp[1]);

      dg[0] = dgst; // + eps*pp[0]; 
      dg[1] = 1-sm - dgst; // + eps*pp[1];

      if(d2g == NULL) return( (1-sm)*pp[1] + gst );  

      d2g[0][0] = d2g[1][1] = eps;
      d2g[0][1] = d2g[1][0] = 0.0;

      return( (1-sm)*pp[1] + gst );   
    }
  else           
    {
      dgst = 1/(4*delp*delp); 
      gst = -1/(4*delp) - sqrt(1-sm); // + (eps/2)*(pp[0]*pp[0] + pp[1]*pp[1]);

      dg[0] = dgst; // + eps*pp[0]; 
      dg[1] = 1-sm - dgst; // + eps*pp[1];

      if(d2g == NULL) return( (1-sm)*pp[1] + gst );  

      d2g[0][0] = d2g[1][1] = -2.0*dgst/delp + eps;
      d2g[0][1] = d2g[1][0] =  2.0*dgst/delp;

      return( (1-sm)*pp[1] + gst );  
    }
}

double gammast2(const double sm, double pp[2], double dg[2])
{
  return( gammast2(sm,pp, dg, NULL) );
}

double coeff2(const double t, const dvector xx,
	      double p0[2], double p1[2], 
	      double  &sm,   double s0[2],
	      double  dq[2], double d2q[2][2], 
	      dvector bb[2], dmatrix kk[2])

{
  dvector c0 = xx - dvector(-1,-1);
  dvector c1 = xx - dvector( 1, 1);

  sm = porosity2(xx);

  gammast2(sm, p0, s0);

  bb[0].zero(); bb[1].zero();

  dvector cc = xx - dvector(0.25,-0.25);
  
  dmatrix km = 2.5 * fmax(exp(cc.dot(cc))-1.25, 0.0) * dmatrix().identity();

  double k0 = 2.0 * s0[0] * fmax(0.0, s0[0]-0.1*(1-sm));
  double k1 = 0.5 * s0[1] * fmax(0.0, s0[1]-0.1*(1-sm));

  double qq = 0.0;
  dq[0] = dq[1] = 0.0;
  d2q[0][0] = d2q[1][0] = d2q[0][1] = d2q[1][1] = 0.0;

  if(c0.dot(c0) < 1.0/64.0)   // injection well
    {
      dq[0] = p1[0] - 60;
      //      dq[1] = p1[1];
      
      qq = 0.5 * (dq[0]*dq[0] + dq[1]*dq[1]);

      d2q[0][0] = 1.0;
      //      d2q[1][1] = 1.0;

      //km.identity();
      //k0 += 0.25;   
      //k1 += 0.25;   
    }

  if(c1.dot(c1) < 1.0/64.0)   // production well
    {
      dq[0] = p1[0];
      dq[1] = p1[1];   // mean pressure zero on production well
      
      qq = 0.5 * (dq[0]*dq[0] + dq[1]*dq[1]);

      d2q[0][0] = d2q[1][1] = 1.0;

      //km.identity();
      //k0 += 0.25;   
      //k1 += 0.25;   
    }

  kk[0] = k0 * km;
  kk[1] = k1 * km;

  return(qq);
}
